/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: Catalog.c++
 *
 * Description: A component that provides a stack of forms each with a
 *	tab to select a particular form.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include <stdio.h>
#include <Xm/Form.h>
#include <Vk/VkWidgetList.h>
#include "common.h"
#include "Catalog.h"


/**************************************************************************
 *
 * Function: Catalog
 *
 * Description: Constructor for the class.
 *
 * Parameters: 
 *	name (I) - name of the class instance
 *	parent (I) - parent widget for the instance
 *
 * Return: none
 *
 **************************************************************************/

Catalog::Catalog(const char *name, Widget parent)
    :VkComponent(name)
{
    DEFARGS(10);

    /*
     * Instantiate a drawing area as our base widget.
     */
    _baseWidget = XtCreateWidget(_name, xmFormWidgetClass, parent, NULL, 0);

    /*
     * Install a destruction handler
     */
    installDestroyHandler();

    /*
     * Create the component list
     */
    list = new VkComponentList();

    /*
     * Create a deck
     */
    deck = new Deck("deck", _baseWidget);
    deck->show();
    STARTARGS;
    SETARG(XmNtopAttachment, XmATTACH_FORM);
    SETARG(XmNtopOffset, 2);
    SETARG(XmNleftAttachment, XmATTACH_FORM);
    SETARG(XmNleftOffset, 2);
    SETARG(XmNrightAttachment, XmATTACH_FORM);
    SETARG(XmNrightOffset, 2);
    XtSetValues(*deck, ARGLIST);

    /*
     * Create a tab panel
     */
    tabs = new VkTabPanel("tabs", _baseWidget);
    tabs->show();
    STARTARGS;
    SETARG(XmNtopAttachment, XmATTACH_WIDGET);
    SETARG(XmNtopWidget, deck->baseWidget());
    SETARG(XmNtopOffset, 2);
    SETARG(XmNleftAttachment, XmATTACH_FORM);
    SETARG(XmNrightAttachment, XmATTACH_FORM);
    SETARG(XmNbottomAttachment, XmATTACH_FORM);
    XtSetValues(*tabs, ARGLIST);

    /*
     * Add the tab selection callback
     */
    tabs->addCallback(VkTabPanel::tabSelectCallback, this,
				(VkCallbackMethod)&Catalog::tabSelectedCB);
}


/**************************************************************************
 *
 * Function: ~Catalog
 *
 * Description: Destructor for the class.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

Catalog::~Catalog()
{
}


/**************************************************************************
 *
 * Function: className
 *
 * Description: Returns the name of the class.
 *
 * Parameters: none
 *
 * Return: Pointer to class name string.
 *
 **************************************************************************/

const char* Catalog::className()
{
    return "Catalog";
}


/**************************************************************************
 *
 * Function: addView
 *
 * Description: Adds a component to the catalog.
 *
 * Parameters: 
 *	component (I) - component to add
 *	label (I) - label for the tab
 *
 * Return: none
 *
 **************************************************************************/

void Catalog::addView(VkComponent *component, char *label)
{
    /*
     * Add the component to the list
     */
    list->add(component);

    /*
     * Add the component to the deck
     */
    deck->addView(component);

    /*
     * Add the tab
     */
    int ind = tabs->addTab(label, (void*)component);
    tabs->selectTab(ind);
}


/**************************************************************************
 *
 * Function: removeView
 *
 * Description: Removes a component from the catalog.
 *
 * Parameters: 
 *	component (I) - component to remove
 *
 * Return: none
 *
 **************************************************************************/

void Catalog::removeView(VkComponent *component)
{
    int i;
    VkComponent *tabComp;
    char *label;

    /*
     * Remove the component from the list
     */
    list->remove(component);

    /*
     * Remove the tab
     */
    for (i = 0; i < tabs->size(); i++) {
	tabs->getTab(i, &label, (void**)&tabComp);
	if (tabComp == component) {
	    tabs->removeTab(i);
	    break;
	}
    }

    /*
     * Remove the component from the deck
     */
    deck->removeView(component);
}


/**************************************************************************
 *
 * Function: afterRealizeHook
 *
 * Description: Works around a double buffering bug in the VkTabPanel.
 *	The problem is that since we defer showing the main window until
 *	after we get a vi window mapped, the tab panel gets confused and
 *	misses a swap buffer. This forces the swap buffer.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void Catalog::afterRealizeHook()
{
    tabs->selectTab(tabs->selectedTab());
}


/**************************************************************************
 *
 * Function: tabSelectedCB
 *
 * Description: Called when a tab is selected.
 *
 * Parameters: 
 *	obj (I) - component that triggered the callback
 *	clientData (I) - not used
 *	callData (I) - callback data
 *
 * Return: none
 *
 **************************************************************************/

void Catalog::tabSelectedCB(VkCallbackObject*, void*, void *callData)
{
    VkTabCallbackStruct *cb = (VkTabCallbackStruct*)callData;

    /*
     * If there are no more items in catalog, don't do anything
     */
    if (list->size() == 0)
        return;

    /*
     * Pop the appropriate component to the top of the deck
     */
    deck->pop((VkComponent*)cb->clientData);
}


/**************************************************************************
 *
 * Function: getCurrent
 *
 * Description: Returns the currently selected entry.
 *
 * Parameters: none
 *
 * Return: Currently selected entry of NULL if no entries in catalog.
 *
 **************************************************************************/

VkComponent* Catalog::getCurrent()
{
    if (list->size() == 0)
	return NULL;

    return (*list)[tabs->selectedTab()];
}


/**************************************************************************
 *
 * Function: setCurrent
 *
 * Description: Sets the specified component as current (i.e. pops the
 *	specified component to the top of the stack).
 *
 * Parameters: 
 *	comp (I) - component to make current
 *
 * Return: none
 *
 **************************************************************************/

void Catalog::setCurrent(VkComponent *comp)
{
    int i;
    char *label;
    VkComponent *existingComp;

    for (i = 0; i < tabs->size(); i++) {
	if (tabs->getTab(i, &label, (void**)&existingComp)) {
	    if (existingComp == comp)
		tabs->selectTab(i);
	}
    }
}


/**************************************************************************
 *
 * Function: setTabLabel
 *
 * Description: Set the specified string as the label for the currently
 *	selected tab.
 *
 * Parameters: 
 *	label (I) - new text for tab
 *
 * Return: none
 *
 **************************************************************************/

void Catalog::setTabLabel(char *label)
{
    tabs->setTabLabel(tabs->selectedTab(), label);
}
