/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: Deck.c++
 *
 * Description: Stacked deck component. Multiple widgets may be stacked
 *	on this component and poped or pushed for visibility. This component
 *	is modeled after the ViewKit VkDeck component but it does not use
 *	the Form as the baseWidget. This is because the Motif Form widget
 *	does not really know how to handle stacked widget during resize.
 *	In addition, this component adds functionality to the original
 *	ViewKit deck.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include "common.h"
#include "Deck.h"
#include <Vk/VkWidgetList.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>


/**************************************************************************
 *
 * Function: Deck
 *
 * Description: Constructor for the Deck component class. Do the usual
 *	component things like create the base widget and init class state
 *	variables.
 *
 * Parameters: 
 *	name (I) - name of the component
 *	parent (I) - component's parent widget
 *
 * Return: none
 *
 **************************************************************************/

Deck::Deck(const char *name, Widget parent): VkComponent(name)
{
    DEFARGS(10);
    Pixel bg;

    /*
     * Init state variables
     */
    _children = NULL;
    _current = -1;

    /*
     * We want our background to be the same as our parent's
     */
    STARTARGS;
    SETARG(XmNbackground, &bg);
    XtGetValues(parent, ARGLIST);

    /*
     * Create our base widget and install the destroy handler
     */
    STARTARGS;
    SETARG(XmNbackground, bg);
    SETARG(XmNmarginWidth, 1);
    SETARG(XmNmarginHeight, 1);
    _baseWidget = XmCreateDrawingArea(parent, _name, ARGLIST);
    installDestroyHandler();
}


/**************************************************************************
 *
 * Function: ~Deck
 *
 * Description: Destructor for the Deck class. Deletes all the deck's
 *	children.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

Deck::~Deck()
{
    delete _children;
}


/**************************************************************************
 *
 * Function: className
 *
 * Description: Returns the name of the class.
 *
 * Parameters: none
 *
 * Return: Pointer to class name string.
 *
 **************************************************************************/

const char* Deck::className()
{
    return "Deck";
}


/**************************************************************************
 *
 * Function: afterRealizeHook
 *
 * Description: Called after the component is realized so that we can
 *	ensure that the current widget is at the top.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void Deck::afterRealizeHook()
{
    if (_current >= 0) {
	popWidget((*_children)[_current]);
    }

}


/**************************************************************************
 *
 * Function: addView
 *
 * Description: Adds either a widget or component child to the deck.
 *
 * Parameters: 
 *	child (I) - widget to be added to the deck
 *	comp (I) - component to be added to the deck
 *
 * Return: none
 *
 **************************************************************************/

void Deck::addView(Widget child)
{
    if (!_children)
	_children = new VkWidgetList();

    _children->add(child);
}

void Deck::addView(VkComponent *comp)
{
    addView(comp->baseWidget());
}


/**************************************************************************
 *
 * Function: removeView
 *
 * Description: Removes the specified component from the deck.
 *
 * Parameters:
 *      component (I) - component to remove from the deck.
 *
 * Return: none
 *
 **************************************************************************/

void Deck::removeView(VkComponent *component)
{
    int i;

    /*
     * If this is the last child in the deck, set the state accordingly
     */
    if (_children->size() == 1)
        _current = -1;

    /*
     * Find the component's base widget in the deck's child list
     */
    for (i = 0; i < _children->size(); i++) {
        if ((*_children)[i] == component->baseWidget()) {
            _children->remove(component->baseWidget());
            return;
        }
    }
}


/**************************************************************************
 *
 * Function: popWidget
 *
 * Description: Pops the specified child widget to the top of the deck.
 *	The poping is done by raising the widget's window to the top of
 *	of the window stack.
 *
 * Parameters: 
 *	w (I) - widget child of Deck to pop to the top. If NULL, the
 *		current widget is popped to the top.
 *
 * Return: none
 *
 **************************************************************************/

void Deck::popWidget(Widget w)
{
    if (!w) {
	_current++;

	if (_current >= _children->size())
	    _current = 0;

	w = (*_children)[_current];
    } else {
	for (int i = 0; i < _children->size(); i++)
	    if ((*_children)[i] == w)
		_current = i;
    }

    if (XtIsRealized(w))
	XMapRaised(XtDisplay(_baseWidget), XtWindow(w));
}


/**************************************************************************
 *
 * Function: pushWidget
 *
 * Description: Pushes the specified child widget to the down the stack.
 *
 * Parameters: 
 *	w (I) - widget child of Deck to push
 *
 * Return: none
 *
 **************************************************************************/

void Deck::pushWidget(Widget w)
{
    if (!w) {
	_current--;

	if (_current < 0)
	    _current = _children->size() - 1;

	w = (*_children)[_current];
    } else {
	for (int i = _children->size(); i > 0; i--)
	    if ((*_children)[i] == w)
		_current = i;
    }

    if (XtIsRealized(w))
	XMapRaised(XtDisplay(_baseWidget), XtWindow(w));
}


/**************************************************************************
 *
 * Function: pop
 *
 * Description: Pops the deck.
 *
 * Parameters: 
 *	w (I) - widget to pop to the top of the deck
 *	comp (I) - component to pop to the top of the deck
 *
 * Return: none
 *
 **************************************************************************/

void Deck::pop(Widget w)
{
    popWidget(w);
}

void Deck::pop()
{
    pop((Widget) NULL);
}

void Deck::pop(VkComponent* comp)
{
    pop(comp->baseWidget());
}


/**************************************************************************
 *
 * Function: push
 *
 * Description: Pushes the deck.
 *
 * Parameters: 
 *	w (I) - widget to push
 *	comp (I) - component to push
 *
 * Return: none
 *
 **************************************************************************/

void Deck::push()
{
    push((Widget) NULL);
}

void Deck::push(Widget w)
{
    pushWidget(w);
}

void Deck::push(VkComponent* comp)
{
    push(comp->baseWidget());
}
