/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: MainWindow.c++
 *
 * Description: Main window program file for the graphical vi program, gvi.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>
#include <Xm/MessageB.h>
#include <Xm/Form.h>
#include <Xm/List.h>
#include <Xm/Text.h>
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>
#include <Vk/VkMenuBar.h>
#include <Vk/VkResource.h>
#include "safeMemory.h"
#include "common.h"
#include "MainWindow.h"


// No filename tab label

static char* noFileLabel = "<noname>";


// Application main menu

VkMenuDesc MainWindow::fileMenu[] = {
    { SEPARATOR },
    { ACTION, "new", MainWindow::newCB },
    { ACTION, "open", MainWindow::openCB },
    { ACTION, "openRecent", MainWindow::openRecentCB },
    { SEPARATOR },
    { ACTION, "normalClose", MainWindow::normalCloseCB },
    { ACTION, "saveClose", MainWindow::saveCloseCB },
    { ACTION, "forceClose", MainWindow::forceCloseCB },
    { SEPARATOR },
    { ACTION, "quit", MainWindow::quitCB },
    { END }
};

VkMenuDesc MainWindow::selectedMenu[] = {
    { SEPARATOR },
    { ACTION, "changeTabLabel", MainWindow::changeTabLabelCB },
    { END }
};

VkMenuDesc MainWindow::mainMenu[] = {
    { SUBMENU, "fileMenu", NULL, fileMenu },
    { SUBMENU, "selectedMenu", NULL, selectedMenu },
    { END }
};


/**************************************************************************
 *
 * Function: MainWindow
 *
 * Description: Constructor for the giv main window class
 *
 * Parameters: 
 *	name (I) - name of window instance
 *
 * Return: none
 *
 **************************************************************************/

MainWindow::MainWindow(const char *name, ArgList args,
			Cardinal argCount): VkWindow(name, args, argCount)
{
    XWindowAttributes xwa;
    ViWsh *viwsh;
    DEFARGS(10);

    /*
     * Initialize state variables
     */
    fileSelectionDialog = NULL;
    newFileDialog = NULL;
    lastFileDialog = NULL;
    changeTabDialog = NULL;
    recentFilesDialog = NULL;
    curWidth = 0;
    curHeight = 0;
    deleteOnMap = NULL;

    /*
     * Create the Recent Files manager
     */
    recentFiles = new RecentFiles();

    /*
     * Create the deletion pipe. When viwsh's die we want to delete their
     * instances. We do not want to do this in the vi death callback
     * because we would be deallocating storage that the callback needs
     * when it returns. So we write the instance address to the delete pipe
     * and the object is deleted when the event loop resumes.
     */
    if (pipe(deletePipe) == -1) {
	perror("pipe");
	exit(1);
    }
    XtAppAddInput(theApplication->appContext(), deletePipe[0],
			(XtPointer)XtInputReadMask, deleteViCB,
			(XtPointer)this);

    /*
     * Create the app menu bar
     */
    menuBar = new VkMenuBar(mainMenu, (XtPointer)this);
    setMenuBar(menuBar);

    /*
     * We need to process window substructure change events to catch
     * the wsh mapping events. We also want size change events for the
     * app window.
     */
    XGetWindowAttributes(XtDisplay(_baseWidget),
                RootWindowOfScreen(XtScreen(_baseWidget)), &xwa);
    XSelectInput(XtDisplay(_baseWidget),
                RootWindowOfScreen(XtScreen(_baseWidget)),
                xwa.your_event_mask | SubstructureNotifyMask);
    XSync(XtDisplay(_baseWidget), False);

    /*
     * Create a top level form to put everything in
     */
    Widget form = XtCreateManagedWidget("form", xmFormWidgetClass, *this,
							NULL, 0);
    addView(form);

    /*
     * Create the catalog for the vi windows
     */
    catalog = new Catalog("catalog", form);
    catalog->show();
    STARTARGS;
    SETARG(XmNtopAttachment, XmATTACH_FORM);
    SETARG(XmNbottomAttachment, XmATTACH_FORM);
    SETARG(XmNleftAttachment, XmATTACH_FORM);
    SETARG(XmNrightAttachment, XmATTACH_FORM);
    XtSetValues(catalog->baseWidget(), ARGLIST);

    /*
     * Create a vi window for each filename on the command line. The
     * first vi is remmebered.
     */
    if (theApplication->argc() > 1) {
        viwsh = createViWsh(theApplication->argv(1), True);
        catalog->addView(viwsh, basename(theApplication->argv(1)));
    } else {
        viwsh = createViWsh(NULL, True);
        catalog->addView(viwsh, noFileLabel);
    }
    viwsh->addCallback(ViWsh::mappedCallback, this,
			(VkCallbackMethod)&MainWindow::viMappedCB);

    /*
     * Files after the first are brought up in the background
     */
    if (theApplication->argc() > 2) {
	ViBackground *backProc = new ViBackground();
	backProc->start();
    }
}


/**************************************************************************
 *
 * Function: ~MainWindow
 *
 * Description: Destructor for the application main window class
 *
 * Parameters: name
 *
 * Return: none
 *
 **************************************************************************/

MainWindow::~MainWindow()
{
}


/**************************************************************************
 *
 * Function: className
 *
 * Description: Provides the name of the class
 *
 * Parameters: none
 *
 * Return: Name of the class.
 *
 **************************************************************************/

const char* MainWindow::className()
{
    return "MainWindow";
}


/**************************************************************************
 *
 * Function: viMappedCB
 *
 * Description: Called when a vi/wsh window maps.
 *
 * Parameters: 
 *	obj (I) - component that triggered the callback
 *	clientData (I) - not used
 *	callData (I) - window ID of the mapped viwsh window
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::viMappedCB(VkCallbackObject*, void*, void* callData)
{
    Window wid = (Window)callData;
    XWindowAttributes xwa;
    XSizeHints hints;
    long flags;
    int minWidth = -1, minHeight = -1;
    DEFARGS(5);

    show();

    /*
     * Remember the initial width and height so that later we can compute
     * resize deltas.
     */
    STARTARGS;
    SETARG(XmNwidth, &curWidth);
    SETARG(XmNheight, &curHeight);
    XtGetValues(_baseWidget, ARGLIST);

    /*
     * Set a minimum size on the app.
     */
    XGetWMNormalHints(XtDisplay(_baseWidget), wid, &hints, &flags);
    XGetWindowAttributes(XtDisplay(_baseWidget), wid, &xwa);
    if (flags & PResizeInc) {
	minWidth = curWidth - xwa.width + 40 * hints.width_inc;
	minHeight = curHeight - xwa.height + 5 * hints.height_inc;
    }
    else if (flags & PMinSize) {
	minWidth = 10 * (curWidth - xwa.width + hints.min_width);
	minHeight = 10 * (curHeight - xwa.height + hints.min_height);
    } else {
	minWidth = 300;
	minHeight = 300;
    }
    STARTARGS;
    SETARG(XmNminWidth, minWidth);
    SETARG(XmNminHeight, minHeight);
    XtSetValues(_baseWidget, ARGLIST);
}


/**************************************************************************
 *
 * Function: viDestroyedCB
 *
 * Description: Called when a vi/wsh window process is destroyed.
 *
 * Parameters: 
 *	obj (I) - component that triggered the callback
 *	clientData (I) - not used
 *	callData (I) - callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::viDestroyedCB(VkCallbackObject *obj, void*, void*)
{
    DEFARGS(10);

    /*
     * If this is the last viwsh we put up a dialog to ask the
     * user if they want to quit the app or have a <noname> viwsh.
     * If this is not the last viwsh we remove the view and send the
     * message to delete the viwsh.
     */
    if (catalog->size() == 1) {
	/*
	 * Clear the tab on the empty view
	 */
	catalog->setTabLabel("");

        /*
         * Indicate a delayed delete so that the TabPanel is never without
         * a viwsh. We do not delete the current one until the new one maps.
         * This is all because TabPanel has bugs and does not like to be
	 * emptied. We set this now in case the user chooses to conitnue
	 * in the app.
         */
        deleteOnMap = (ViWsh*)obj;

	/*
	 * Create the quit app question dialog, if needed
	 */
	if (!lastFileDialog) {
	    STARTARGS;
	    SETARG(XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
	    lastFileDialog = XmCreateMessageDialog(baseWidget(),
					"lastFileDialog", ARGLIST);
	    STARTARGS;		// Motif only sets the symbol if we XtSetValues
	    SETARG(XmNdialogType, XmDIALOG_QUESTION);
	    XtSetValues(lastFileDialog, ARGLIST);
	    XtUnmanageChild(XmMessageBoxGetChild(lastFileDialog,
					XmDIALOG_HELP_BUTTON));
	    XtAddCallback(lastFileDialog, XmNokCallback, quitCB, this);
	    XtAddCallback(lastFileDialog, XmNcancelCallback,
					lastFileClosedCB, this);
    	}

	/*
	 * Show the dialog
	 */
        XtManageChild(lastFileDialog);
    } else {
        catalog->removeView((ViWsh*)obj);
        write(deletePipe[1], &obj, sizeof(void*));
    }
}


/**************************************************************************
 *
 * Function: deleteOnMapCB
 *
 * Description: Called when a viwsh window maps to check whether there is
 *	a viwsh object whose deletion is waiting pending this mapping.
 *
 * Parameters: 
 *	obj (I) - component that triggered the callback
 *	clientData (I) - not used
 *	callData (I) - callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::deleteOnMapCB(VkCallbackObject*, void*, void*)
{
    /*
     * If there is a viwsh waiting to be deleted for another viwsh to
     * map, delete it.
     */
    if (deleteOnMap) {
        write(deletePipe[1], &deleteOnMap, sizeof(void*));
	deleteOnMap = NULL;
    }
}


/**************************************************************************
 *
 * Function: deleteViCB
 *
 * Description: Called when a request to delete an instance of a ViWsh
 *	is received over the deletePipe.
 *
 * Parameters: 
 *	client (I) - class instance pointer
 *	fdp (I) - file destriptor of the read end of the delete pipe
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::deleteViCB(XtPointer, int *fdp, XtInputId*)
{
    ViWsh *obj;

    if (read(*fdp, &obj, sizeof(void*)) != sizeof(void*)) {
	perror("read");
	return;
    }
    delete obj;
}


/**************************************************************************
 *
 * Function: afterRealizeHook
 *
 * Description: Called after the main window gets a window assigned. The
 *	job of this function is to register interest in the ConfigureNotify
 *	event so that we may communicate resizing of the app to the ViWsh
 *	window.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::afterRealizeHook()
{
    /*
     * We want the ConfigureNotify event to tell use when we have been
     * resized.
     */
    XtAddEventHandler(_baseWidget, StructureNotifyMask, False,
		  (XtEventHandler)MainWindow::resizeCB, (XtPointer)this);
}


/**************************************************************************
 *
 * Function: resizeCB
 *
 * Description: Called by Xt when the main window is resized in some way.
 *	We simply call our instance resize handler function.
 *
 * Parameters: 
 *	client (I) - class instance pointer
 *	event (I) - X event that triggered the call
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::resizeCB(Widget, XtPointer client, XEvent *event)
{
    if (event->type == ConfigureNotify)
	((MainWindow*)client)->resizeHandler();
}


/**************************************************************************
 *
 * Function: handleRawEvent
 *
 * Description: Event handler for raw X events of interest to the ViWsh
 *	class.
 *
 * Parameters: 
 *	event (I) - raw X event
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::handleRawEvent(XEvent *event)
{
    int i;

    /*
     * If this is a MapNotify event, pass the event to the ViWsh
     * class instances. Note that we only pass MapNotify events
     * for override redirect windows.
     */
    if (event->type == MapNotify && event->xmap.override_redirect &&
		event->xmap.window) {
	for (i = 0; i < catalog->size(); i++) {
	    ((ViWsh*)((*catalog)[i]))->handleMapping(event->xmap.window);
	}
    }
}


/**************************************************************************
 *
 * Function: newCB
 *
 * Description: Create a new file callback.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::newCB(Widget, XtPointer client, XtPointer)
{
    DEFARGS(10);
    MainWindow *win = (MainWindow*)client;

    if (!win->newFileDialog) {
	STARTARGS;
	SETARG(XmNautoUnmanage, True);
	SETARG(XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
	win->newFileDialog = XmCreatePromptDialog(win->baseWidget(),
					"newFileDialog", ARGLIST);
	XtUnmanageChild(XmSelectionBoxGetChild(win->newFileDialog,
					XmDIALOG_HELP_BUTTON));
	XtAddCallback(XmSelectionBoxGetChild(win->newFileDialog,
					XmDIALOG_TEXT),
					XmNmodifyVerifyCallback,
					filenameVerifyCB, win);
	XtAddCallback(win->newFileDialog, XmNokCallback,
					newFileSelectedCB, win);
    }
    XtManageChild(win->newFileDialog);
}


/**************************************************************************
 *
 * Function: openCB
 *
 * Description: Open a file callback.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::openCB(Widget, XtPointer client, XtPointer)
{
    DEFARGS(10);
    MainWindow *win = (MainWindow*)client;

    if (!win->fileSelectionDialog) {
	STARTARGS;
	SETARG(XmNautoUnmanage, True);
	win->fileSelectionDialog =
			XmCreateFileSelectionDialog(win->baseWidget(),
					"openFileDialog", ARGLIST);
	XtUnmanageChild(XmFileSelectionBoxGetChild(win->fileSelectionDialog,
					XmDIALOG_HELP_BUTTON));
	XtAddCallback(win->fileSelectionDialog, XmNokCallback,
					fileSelectedCB, win);
    }
    XtManageChild(win->fileSelectionDialog);
}


/**************************************************************************
 *
 * Function: openRecentCB
 *
 * Description: Open recent files callback.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::openRecentCB(Widget, XtPointer client, XtPointer)
{
    int numFiles, i;
    char **filenames;
    DEFARGS(10);
    MainWindow *win = (MainWindow*)client;

    if (!win->recentFilesDialog) {
	STARTARGS;
	SETARG(XmNautoUnmanage, True);
	win->recentFilesDialog = 
			XmCreateSelectionDialog(win->baseWidget(),
					"openRecentDialog", ARGLIST);
	XtUnmanageChild(XmFileSelectionBoxGetChild(win->recentFilesDialog,
					XmDIALOG_HELP_BUTTON));
	XtUnmanageChild(XmFileSelectionBoxGetChild(win->recentFilesDialog,
					XmDIALOG_TEXT));
	XtUnmanageChild(XmFileSelectionBoxGetChild(win->recentFilesDialog,
					XmDIALOG_SELECTION_LABEL));
	XtAddCallback(win->recentFilesDialog, XmNokCallback,
					recentFilesSelectedCB, win);
	XtAddCallback(win->recentFilesDialog, XmNapplyCallback,
					recentFilesSelectedCB, win);
	STARTARGS;
	SETARG(XmNselectionPolicy, XmEXTENDED_SELECT);
	XtSetValues(XmFileSelectionBoxGetChild(win->recentFilesDialog,
			XmDIALOG_LIST), ARGLIST);
    }

    /*
     * Put the recently opened filenames in the list
     */
    Widget list = XmFileSelectionBoxGetChild(win->recentFilesDialog,
					XmDIALOG_LIST);
    XmListDeleteAllItems(list);
    if ((numFiles = win->recentFiles->getFilenames(&filenames)) != 0) {
	XmString *strs = (XmString*)safeMalloc(numFiles * sizeof(XmString));
	for (i = 0; i < numFiles; i++)
	    strs[i] = XmStringCreateLocalized(filenames[i]);
	XmListAddItems(list, strs, numFiles, 1);
	for (i = 0; i < numFiles; i++)
	    XmStringFree(strs[i]);
	free(strs);
	win->recentFiles->freeFilenames(filenames);
    }

    XtManageChild(win->recentFilesDialog);
}


/**************************************************************************
 *
 * Function: normalCloseCB
 *
 * Description: Sends a close command ('q') to the vi window.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::normalCloseCB(Widget, XtPointer client, XtPointer)
{
    MainWindow *win = (MainWindow*)client;
    ViWsh *viwsh;
    
    if ((viwsh = (ViWsh*)(win->catalog->getCurrent())) == NULL)
	return;
    viwsh->sendString("\033:q\n");
}


/**************************************************************************
 *
 * Function: saveCloseCB
 *
 * Description: Sends a save and close command ('ZZ') to the vi window.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::saveCloseCB(Widget, XtPointer client, XtPointer)
{
    MainWindow *win = (MainWindow*)client;
    ViWsh *viwsh;
    
    if ((viwsh = (ViWsh*)(win->catalog->getCurrent())) == NULL)
	return;
    viwsh->sendString("\033ZZ\n");
}


/**************************************************************************
 *
 * Function: forceCloseCB
 *
 * Description: Sends a force close command ('q!') to the vi window.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::forceCloseCB(Widget, XtPointer client, XtPointer)
{
    MainWindow *win = (MainWindow*)client;
    ViWsh *viwsh;
    
    if ((viwsh = (ViWsh*)(win->catalog->getCurrent())) == NULL)
	return;
    viwsh->sendString("\033:q!\n");
}


/**************************************************************************
 *
 * Function: lastFileClosedCB
 *
 * Description: Called when the user choses to not exit the app after the
 *	last file has been closed. We run another instance of a viwsh
 *	with a noname file.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::lastFileClosedCB(Widget, XtPointer client, XtPointer)
{
    MainWindow *win = (MainWindow*)client;

    ViWsh *viwsh = win->createViWsh(NULL, True);
    win->catalog->addView(viwsh, noFileLabel);
    win->catalog->removeView(win->deleteOnMap);
}


/**************************************************************************
 *
 * Function: changeTabLabelCB
 *
 * Description: Called when the user wants to change the text of the
 *	currently selected tab.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::changeTabLabelCB(Widget, XtPointer client, XtPointer)
{
    MainWindow *win = (MainWindow*)client;
    DEFARGS(5);

    if (!win->changeTabDialog) {
	STARTARGS;
	SETARG(XmNautoUnmanage, True);
	win->changeTabDialog = XmCreatePromptDialog(win->baseWidget(),
					"changeTabDialog", ARGLIST);
	XtManageChild(XmSelectionBoxGetChild(win->changeTabDialog,
					XmDIALOG_APPLY_BUTTON));
	XtUnmanageChild(XmSelectionBoxGetChild(win->changeTabDialog,
					XmDIALOG_HELP_BUTTON));
	XtAddCallback(win->changeTabDialog, XmNokCallback,
					tabLabelSelectedCB, win);
	XtAddCallback(win->changeTabDialog, XmNapplyCallback,
					tabLabelSelectedCB, win);
    }
    XtManageChild(win->changeTabDialog);
}


/**************************************************************************
 *
 * Function: tabLabelSelectedCB
 *
 * Description: Called when a new tab label is selected from the Change
 *	Tab Label dialog.
 *
 * Parameters: 
 *	w (I) - widget that triggered callback
 *	client (I) - main window instance pointer
 *	callData (I) - tab label text callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::tabLabelSelectedCB(Widget, XtPointer client,
							XtPointer callData)
{
    char *tabLabel;
    FileInfo info;

    XmSelectionBoxCallbackStruct *cb = (XmSelectionBoxCallbackStruct*)callData;

    MainWindow *win = (MainWindow*)(client);
    ViWsh *viwsh = (ViWsh*)win->catalog->getCurrent();

    /*
     * Set the tab label in the cvatalog
     */
    char *pathname = viwsh->getPathname();
    if (cb->reason == XmCR_OK) {
        if (!XmStringGetLtoR(cb->value, XmFONTLIST_DEFAULT_TAG, &tabLabel))
	    return;
        info.label = tabLabel;
    } else if (cb->reason == XmCR_APPLY) {
	if (pathname == NULL)
	    tabLabel = XtNewString(noFileLabel);
	else
	    tabLabel = XtNewString(basename(pathname));
	info.label = NULL;
	XtUnmanageChild(win->changeTabDialog);
    } else
	return;
    win->catalog->setTabLabel(tabLabel);

    /*
     * Save the tab label in the disk file
     */
    if (pathname)
        win->recentFiles->update(&info);

    XtFree(tabLabel);
}


/**************************************************************************
 *
 * Function: newFileSelectedCB
 *
 * Description: Called when a new filename is selected from the New dialog.
 *
 * Parameters: 
 *	w (I) - widget that triggered callback
 *	client (I) - main window instance pointer
 *	callData (I) - filename text callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::newFileSelectedCB(Widget, XtPointer client,
							XtPointer callData)
{
    char *fname;

    XmSelectionBoxCallbackStruct *cb = (XmSelectionBoxCallbackStruct*)callData;

    if (cb->reason != XmCR_OK)
	return;
    if (!XmStringGetLtoR(cb->value, XmFONTLIST_DEFAULT_TAG, &fname))
	return;
    if ((*fname == '\0') || (strcmp(fname, ".") == 0) ||
					(strcmp(fname, "..") == 0)) {
        XtFree(fname);
	return;
    }

    MainWindow *win = (MainWindow*)(client);
    win->openFile(fname, True);

    XtFree(fname);
}


/**************************************************************************
 *
 * Function: filenameVerifyCB
 *
 * Description: Called when new text is to be inserted or deleted in the
 *	new filename entry text field. This callback helps ensure that a
 *	valid filename is entered into the new filename entry field.
 *
 * Parameters: 
 *	w (I) - widget that triggered callback
 *	client (I) - main window instance pointer
 *	callData (I) - text field modify verify callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::filenameVerifyCB(Widget, XtPointer, XtPointer callData)
{
    XmTextVerifyCallbackStruct *cb = (XmTextVerifyCallbackStruct*)callData;
    int i;

    /*
     * Assume we will do it
     */
    cb->doit = True;

    if (cb->text->ptr == NULL)
	return;

    for (i = 0; i < cb->text->length; i++) {
	if (cb->text->ptr[i] == ' ') {
    	    cb->doit = False;
	    return;
	}
    }
}


/**************************************************************************
 *
 * Function: fileSelectedCB
 *
 * Description: Called when a file is selected from the Open dialog.
 *
 * Parameters: 
 *	w (I) - widget that triggered callback
 *	client (I) - main window instance pointer
 *	callData (I) - file selection callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::fileSelectedCB(Widget, XtPointer client, XtPointer callData)
{
    char *fname;

    XmFileSelectionBoxCallbackStruct *cb =
				(XmFileSelectionBoxCallbackStruct*)callData;
    MainWindow *win = (MainWindow*)(client);

    if (cb->reason != XmCR_OK)
	return;
    if (!XmStringGetLtoR(cb->value, XmFONTLIST_DEFAULT_TAG, &fname))
	return;
    if (win->checkFileExists(fname) == False) {
        XtFree(fname);
	return;
    }

    win->openFile(fname, True);

    XtFree(fname);
}


/**************************************************************************
 *
 * Function: recentFilesSelectedCB
 *
 * Description: Called when a list of recent files has been selected
 *
 * Parameters: 
 *	w (I) - widget that triggered callback
 *	client (I) - main window instance pointer
 *	callData (I) - file selection callback data
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::recentFilesSelectedCB(Widget, XtPointer client, XtPointer)
{
    DEFARGS(10);
    XmStringTable strs;
    int i, numStr;
    char *fname, *label;

    MainWindow *win = (MainWindow*)(client);
    STARTARGS;
    SETARG(XmNselectedItemCount, &numStr);
    SETARG(XmNselectedItems, &strs);
    XtGetValues(XmFileSelectionBoxGetChild(win->recentFilesDialog,
						XmDIALOG_LIST), ARGLIST);
    for (i = 0; i < numStr; i++) {
        if (!XmStringGetLtoR(strs[i], XmFONTLIST_DEFAULT_TAG, &fname))
	    continue;
        if (win->checkFileExists(fname) == False) {
            XtFree(fname);
	    continue;
        }

	win->openFile(fname, False);

	label = win->recentFiles->getLabel(fname);
	if (label)
	    win->catalog->setTabLabel(label);

        XtFree(fname);
    }
}


/**************************************************************************
 *
 * Function: destroyDialogCB
 *
 * Description: Called by a dialog that wishes to be destroyed when it is
 *	poped down.
 *
 * Parameters: 
 *	client (I) - dialog instance to be destroyed
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::destroyDialogCB(Widget, XtPointer client, XtPointer)
{
    XtDestroyWidget((Widget)client);
}


/**************************************************************************
 *
 * Function: quitCB
 *
 * Description: Application normal exit callback.
 *
 * Parameters: 
 *	client (I) - main window instance pointer
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::quitCB(Widget, XtPointer client, XtPointer)
{
    ((MainWindow*)client)->quit();
}


/**************************************************************************
 *
 * Function: quit
 *
 * Description: Exits the program with a successful return code.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::quit()
{
    theApplication->quitYourself();
}


/**************************************************************************
 *
 * Function: resizeHandler
 *
 * Description: Determines the size change delta and informs the ViWsh's
 *	about the size change.
 *	
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::resizeHandler()
{
    DEFARGS(10);
    Dimension w, h;
    int i, dw, dh;

    /*
     * Get the new width and height
     */
    STARTARGS;
    SETARG(XmNwidth, &w);
    SETARG(XmNheight, &h);
    XtGetValues(_baseWidget, ARGLIST);

    /*
     * Compute the change in width and height and update the current
     * width and height for next time.
     */
    dw = w - curWidth;
    dh = h - curHeight;
    curWidth = w;
    curHeight = h;

    /*
     * If there was no change don't do anything
     */
    if (dw == 0 && dh == 0)
	return;

    /*
     * Now tell the viwsh's about the new size
     */
    for (i = 0; i < catalog->size(); i++)
	((ViWsh*)((*catalog)[i]))->handleResize(dw, dh);
}


/**************************************************************************
 *
 * Function: createViWsh
 *
 * Description: Convenience function for instantiating a viwsh and doing
 *	the necessary callback wiring and other housekeeping.
 *
 * Parameters: 
 *	filename (I) - name of file to open or NULL if no files
 *	width (I) - width in pixels for the viwsh
 *	height (I) - height in pixels for the viwsh
 *	update (I) - True = update the recent files list
 *
 * Return: Pointer to a viwsh instance.
 *
 **************************************************************************/

ViWsh* MainWindow::createViWsh(char *filename, Boolean update, Dimension width,
					Dimension height)
{
    ViWsh *viwsh;
    FileInfo info;

    /*
     * Create the viwsh
     */
    if (width == 0 || height == 0)
        viwsh = new ViWsh("viWsh", catalog->deckWidget(), filename);
    else
        viwsh = new ViWsh("viWsh", catalog->deckWidget(), filename,
					width, height);
    viwsh->show();
    viwsh->addCallback(ViWsh::mappedCallback, this,
			(VkCallbackMethod)&MainWindow::deleteOnMapCB);
    viwsh->addCallback(ViWsh::destroyedCallback, this,
			(VkCallbackMethod)&MainWindow::viDestroyedCB);

    /*
     * If a filename has been specified, register it with the recent files
     * manager. We query the viwsh for the pathname of the file because it
     * will return an absolute pathname for the file.
     */
    if (filename && update) {
	info.pathname = viwsh->getPathname();
	info.label = NULL;
	recentFiles->update(&info);
    }

    return viwsh;
}


/**************************************************************************
 *
 * Function: checkFileExists
 *
 * Description: Checks the the specified file exists. If not, an error
 *	dialog is displayed.
 *
 * Parameters: 
 *	filename (I) - name of file to check for existence
 *
 * Return: False if file does not exist and True if file does exist.
 *
 **************************************************************************/

Boolean MainWindow::checkFileExists(const char *filename)
{
    struct stat sbuf;
    int rvExists, rvIsFile;
    char *msg, *buffer;
    XmString str;
    Atom destroy_atom;
    DEFARGS(5);

    rvExists = (access(filename, F_OK) == 0) && (stat(filename, &sbuf) == 0);
    rvIsFile = (sbuf.st_mode & S_IFDIR) == 0;

    if (rvExists && rvIsFile)
	return True;

    Widget dialog = XmCreateInformationDialog(baseWidget(), "noFileDialog",
							NULL, 0);
    if (!rvExists) {
	msg = VkGetResource(dialog, "cannotAccessString", "CannotAccessString");
    } else {
	msg = VkGetResource(dialog, "isDirectoryString", "IsDirectoryString");
    }
    buffer = (char*)safeMalloc(strlen(msg) + strlen(filename) + 10);
    sprintf(buffer, msg, filename);
    str = XmStringCreateLocalized(buffer);
    STARTARGS;
    SETARG(XmNmessageString, str);
    XtSetValues(dialog, ARGLIST);
    XmStringFree(str);
    free(buffer);

    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));

    XtAddCallback(dialog, XmNokCallback, destroyDialogCB, dialog);
    XtAddCallback(dialog, XmNcancelCallback, destroyDialogCB, dialog);
    if ((destroy_atom = XmInternAtom(XtDisplay(dialog),
                                        "WM_DELETE_WINDOW", True)) != None) {
        XmAddWMProtocols(XtParent(dialog), &destroy_atom, 1);
        XmAddWMProtocolCallback(XtParent(dialog), destroy_atom,
                                        destroyDialogCB, dialog);
    }

    XtManageChild(dialog);

    return False;
}


/**************************************************************************
 *
 * Function: openFile
 *
 * Description: Supervises the creation of the viwsh and the registering of
 *	it with the catalog. This function is called by the file selection
 *	dialog callbacks.
 *
 * Parameters: 
 *	filename (I) - filename to open
 *	update (I) - True = update the recent files list
 *
 * Return: none
 *
 **************************************************************************/

void MainWindow::openFile(char *filename, Boolean update)
{
    Dimension w, h;
    int i;
    char *pathname, *existingPath;
    ViWsh *viwsh;
    DEFARGS(5);

    /*
     * Check whether there is already a file by this name open. We compare
     * the full path with the full paths of the viwsh's already open. If
     * this is a dup we simply pop the existing one to the top.
     */
    pathname = makeAbsolutePathname(filename);
    for (i = 0; i < catalog->size(); i++) {
	viwsh = (ViWsh*)((*catalog)[i]);
	if ((existingPath = viwsh->getPathname()) != NULL) {
	    if (strcmp(existingPath, pathname) == 0) {
	        catalog->setCurrent(viwsh);
	        XBell(XtDisplay(_baseWidget), 100);
	        free(pathname);
	        return;
	    }
	}
    }
    free(pathname);

    /*
     * Create the viwsh and add it to the catalog
     */
    STARTARGS;
    SETARG(XmNheight, &h);
    SETARG(XmNwidth, &w);
    XtGetValues(catalog->deckWidget(), ARGLIST);
    viwsh = createViWsh(filename, update, w, h);
    catalog->addView(viwsh, basename(filename));
}


/*
 =========================================================================
 */


/**************************************************************************
 *
 * Function: timeSlice
 *
 * Description: Implements the work proce to bring up vi/wsh components
 *	as a background process.
 *
 * Parameters: none
 *
 * Return: True if work completed. False if need to be called again.
 *
 **************************************************************************/

Boolean ViBackground::timeSlice()
{
    if (fileIndex < theApplication->argc()) {
        char *fname = theApplication->argv(fileIndex);
	MainWindow *win = (MainWindow*)(theApplication->mainWindow());
        ViWsh *viwsh = win->createViWsh(fname, True);
        win->catalog->addView(viwsh, basename(fname));
        fileIndex++;
    }

    return (fileIndex >= theApplication->argc())? True: False;
}
