/**************************************************************************
 *
 * 		  Copyright (c)	1993 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: ProgramMsg.c++
 *
 * Description: Handles the formatting and printing of error, warning and
 *	information messages. The messages are normally printed to
 *	stderr but this can be changed but specifying a file pointer
 *	on class instantiation or during execution.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "common.h"
#include "ProgramMsg.h"


/* Program scope global for messages */

ProgramMsg progMsg("");			// Program messages


/**************************************************************************
 *
 * Function: error
 *
 * Description: Displays an error message.
 *
 * Parameters: 
 *	format (I) - printf style format string
 *	... (I) - arguments for the format string
 *
 * Return: none
 *
 **************************************************************************/
/* VARARGS1 */

void ProgramMsg::error(const char *format, ...)
{
    va_list ap;

    /*
     * Print the specific heading
     */
    (void)fprintf(outPtr, "%s Error: ", progName);

    /*
     * Call the generic display routine
     */
    va_start(ap, format);
    dispMsg(format, ap);
    va_end(ap);
}


/**************************************************************************
 *
 * Function: warn
 *
 * Description: Displays a warning message.
 *
 * Parameters: 
 *	format (I) - printf style format string
 *	... (I) - arguments for the format string
 *
 * Return: none
 *
 **************************************************************************/
/* VARARGS1 */

void ProgramMsg::warn(const char *format, ...)
{
    va_list ap;

    /*
     * Print the specific heading
     */
    (void)fprintf(outPtr, "%s Warning: ", progName);

    /*
     * Call the generic display routine
     */
    va_start(ap, format);
    dispMsg(format, ap);
    va_end(ap);
}


/**************************************************************************
 *
 * Function: info
 *
 * Description: Displays an information message.
 *
 * Parameters: 
 *	format (I) - printf style format string
 *	... (I) - arguments for the format string
 *
 * Return: none
 *
 **************************************************************************/
/* VARARGS1 */

void ProgramMsg::info(const char *format, ...)
{
    va_list ap;

    /*
     * Print the specific heading
     */
    (void)fprintf(outPtr, "%s Info: ", progName);

    /*
     * Call the generic display routine
     */
    va_start(ap, format);
    dispMsg(format, ap);
    va_end(ap);
}


#ifdef DEBUG
/**************************************************************************
 *
 * Function: debug
 *
 * Description: Displays a debug message.
 *
 * Parameters: 
 *	format (I) - printf style format string
 *	... (I) - arguments for the format string
 *
 * Return: none
 *
 **************************************************************************/
/* VARARGS1 */

void ProgramMsg::debug(const char *format, ...)
{
    va_list ap;

    /*
     * Print the specific heading
     */
    (void)fprintf(outPtr, "%s DEBUG: ", progName);

    /*
     * Call the generic display routine
     */
    va_start(ap, format);
    dispMsg(format, ap);
    va_end(ap);
}
#endif /* DEBUG */


/**************************************************************************
 *
 * Function: raw
 *
 * Description: Displays a message without the program name or message
 *	type printed.
 *
 * Parameters: 
 *	format (I) - printf style format string
 *	... (I) - arguments for the format string
 *
 * Return: none
 *
 **************************************************************************/
/* VARARGS1 */

void ProgramMsg::raw(const char *format, ...)
{
    va_list ap;

    /*
     * Call the generic display routine
     */
    va_start(ap, format);
    dispMsg(format, ap);
    va_end(ap);
}


/**************************************************************************
 *
 * Function: dispMsg
 *
 * Description: Displays a message.
 *
 * Parameters: 
 *	format (I) - printf style format string
 *	args (I) - arguments for the format string
 *
 * Return: none
 *
 **************************************************************************/

void ProgramMsg::dispMsg(const char *format, va_list args)
{
    char *sptr, spec[10], type, size;
    int len;

    /*
     * Sanity check the input
     */
    ASSERT(format != NULL);

    /*
     * Output the message. We scan along the format string. If
     * the character is not a '%' we simply output it. If the character
     * is a '%' we extract the format specification and output the
     * argument according to that specification.
     */
    while (*format) {
	/*
	 * Plain character
	 */
	if (*format != '%') {
	    (void)fputc((int)*format, outPtr);
	    format++;
	}
	/*
	 * Character is a '%'
	 */
	else {
	    /*
	     * We have '%%' situation
	     */
	    if (format[1] == '%') {
		(void)fputc((int)'%', outPtr);
		format += 2;
	    }
	    /*
	     * We have an output specification
	     */
	    else {
		/*
		 * Get the specification type, size and number of
		 * characters.
		 */
		if ((sptr = strpbrk(format, "diouxXfeEgGcspn")) == NULL) {
		    (void)fputc((int)*format, outPtr);
		    format++;
		    continue;
		}
		len = sptr - format + 1;
		(void)strncpy(spec, format, len);
		spec[len] = '\0';
		type = spec[len-1];
		size = spec[len-2];
		/*
		 * Output the argument according to the spec
		 */
		switch (type) {
		    case 'c':
			(void)fprintf(outPtr, spec, va_arg(args, int));
			break;
		    case 'd':
		    case 'i':
		    case 'o':
		    case 'u':
		    case 'x':
		    case 'X':
			if (size == 'h')
			    (void)fprintf(outPtr, spec, va_arg(args, int));
			else if (size == 'l')
			    (void)fprintf(outPtr, spec, va_arg(args, long));
			else
			    (void)fprintf(outPtr, spec, va_arg(args, int));
			break;
		    case 'e':
		    case 'E':
		    case 'f':
		    case 'g':
		    case 'G':
			(void)fprintf(outPtr, spec, va_arg(args, double));
			break;
		    case 's':
			(void)fprintf(outPtr, spec, va_arg(args, char*));
			break;
		    case 'p':
			(void)fprintf(outPtr, spec, va_arg(args, void*));
			break;
		    case 'n':
			(void)fprintf(outPtr, spec, va_arg(args, int*));
			break;
		}
		format += len;
	    }
	}
    }
    (void)fputc((int)'\n', outPtr);
    (void)fflush(outPtr);
}
