/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: RecentFiles.c++
 *
 * Description: Class that manages the saving and recalling of the names
 *	of the last N most recently opened files.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <Vk/VkApp.h>
#include <Vk/VkSimpleWindow.h>
#include <Vk/VkResource.h>
#include "safeMemory.h"
#include "common.h"
#include "RecentFiles.h"


#define DEF_MAXFILES		"30"	// Default max number of recent files
#define DISKLIST_NAME		".gvifiles"	// Name of disk based file list
#define DISKLIST_VERSION	1	// Disk file version number

#define MAX_FILELINE		(PATH_MAX + 1024)


// To get around a libc and C++ name space collision and not have
// to define BSD_COMPAT

extern "C" {
extern int flock(int, int);
}



/**************************************************************************
 *
 * Function: RecentFiles
 *
 * Description: Constructor for the class. Main job is to initialize the
 *	internal state.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

RecentFiles::RecentFiles()
{
    char *home;
    struct stat sbuf;

    /*
     * Initialize the state
     */
    fileList = NULL;
    maxFiles = (int)VkGetResource((theApplication->mainWindow())->baseWidget(),
					"maxRecentFiles", "MaxRecentFiles",
					XmRCardinal, DEF_MAXFILES);
    if (maxFiles <= 0) {
	fprintf(stderr, "Invalid maxRecentFiles, %d, reset to default, %s\n",
			maxFiles, DEF_MAXFILES);
	maxFiles = atoi(DEF_MAXFILES);
    }

    /*
     * Form the disk based list file pathname
     */
    if ((home = getenv("HOME")) == NULL)
	diskPathname = NULL;
    else {
	diskPathname = (char*)safeMalloc(strlen(home) + strlen(DISKLIST_NAME)
						+ 5);
	sprintf(diskPathname, "%s/%s", home, DISKLIST_NAME);
    }

    /*
     * Read the disk based file list, if any. Also note the mod time
     * so we do not do any unnecessary reads.
     */
    if (stat(diskPathname, &sbuf) == -1)
	diskModTime = 0;
    else {
	diskModTime = sbuf.st_mtime;
        readList();
    }
}


/**************************************************************************
 *
 * Function: ~RecentFiles
 *
 * Description: Destructor for the class. Main job is to deallocate storage
 *	for the file list.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

RecentFiles::~RecentFiles()
{
    if (diskPathname)
	free(diskPathname);

    while (fileList)
	fileList = remove(fileList);
}


/**************************************************************************
 *
 * Function: update
 *
 * Description: Updates the list of most recently opened files. First the
 *	disk version of the list is checked to see if it has been modified
 *	since we last update our list. Then the specified file is added to
 *	the list and the disk cache is updated.
 *
 * Parameters: 
 *	finfo (I) - file information structure
 *
 * Return: none
 *
 **************************************************************************/

void RecentFiles::update(const FileInfo *finfo)
{
    struct stat sbuf;

    /*
     * If the file mod time is later than our mod time we reread the list
     */
    if (stat(diskPathname, &sbuf) != -1) {
	if (sbuf.st_mtime > diskModTime)
	    readList();
    }

    /*
     * Add the pathname to the list
     */
    add(finfo);

    /*
     * Write the file list to the disk and update our mod time.
     */
    writeList();
    if (stat(diskPathname, &sbuf) != -1)
	diskModTime = sbuf.st_mtime;
}


/**************************************************************************
 *
 * Function: getFilenames
 *
 * Description: Returns a copy of the file list to the caller. The list
 *	is copied into an array for the user. The last entry in the
 *	array is a NULL pointer. The user should call freeFilenames when
 *	the list is no longer needed.
 *
 * Parameters: 
 *	filenamesp (O) - pointer to an array of filename strings. The storage
 *			for the array and the strings is allocated by the
 *			function and should be deallocated with a call to
 *			freeFilenames when no longer needed. The array is
 *			NULL terminated so that at least one element will
 *			always be returned.
 *
 * Return: Number of files in the file list (not counting the NULL entry).
 *
 **************************************************************************/

int RecentFiles::getFilenames(char ***filenamesp)
{
    int i;

    /*
     * Count the number of files in the list
     */
    int count = 0;
    FileItem *item = fileList;
    while (item) {
	count++;
	item = item->next;
    }

    /*
     * Make a NULL termiante array based copy of the list
     */
    char **fnames = (char**)safeCalloc((count + 1), sizeof(char*));
    item = fileList;
    for (i = 0, item = fileList; i < count; i++, item = item->next)
	fnames[i] = safeStrdup(item->info.pathname);
    *filenamesp = fnames;

    return count;
}


/**************************************************************************
 *
 * Function: freeFilenames
 *
 * Description: Frees the storage for the file list obtained by a previous
 *	call to getFilenames.
 *
 * Parameters: 
 *	filenames (I) - file list to free. This must be a NULL terminated
 *			array as obtained from getFileList.
 *
 * Return: none
 *
 **************************************************************************/

void RecentFiles::freeFilenames(char **filenames)
{
    ASSERT(filenames != NULL);
    for (int i = 0; filenames[i]; i++)
	free(filenames[i]);
    free(filenames);
}


/**************************************************************************
 *
 * Function: add
 *
 * Description: Adds the specified filename to the list of most recently
 *	opened files. The filename is added to the beginning of the list
 *	and any subsequent duplicate is removed from the list.
 *
 * Parameters: 
 *	finfo (I) - file information to be added to the list
 *
 * Return: none
 *
 **************************************************************************/

void RecentFiles::add(const FileInfo *finfo)
{
    /*
     * Add the infomation to the list
     */
    if (finfo->pathname == NULL)
	return;
    FileItem *item = (FileItem*)safeCalloc(1, sizeof(FileItem));
    item->info.pathname = safeStrdup(finfo->pathname);
    item->info.label = (finfo->label) ? safeStrdup(finfo->label): NULL;
    if (fileList) {
	item->next = fileList;
	fileList->prev = item;
	fileList = item;
    } else {
	fileList = item;
    }

    /*
     * Remove any duplicates later in the list and remove files
     * from the end if we have more than the max allowed number
     * of remembered files.
     */
    int count = 1;
    if (fileList->next) {
        item = fileList->next;
        while (item) {
	    if (strcmp(fileList->info.pathname, item->info.pathname) == 0) {
		item = remove(item);
	    } else {
		if (count == maxFiles)
		    item = remove(item);
		else {
		    count++;
		    item = item->next;
		}
	    }
        }
    }

     #ifdef DEBUG
     printf("num files = %d\n", count);
     dump();
     #endif
}


/**************************************************************************
 *
 * Function: remove
 *
 * Description: Removes the specified item from the list.
 *
 * Parameters: 
 *	item (I) - item to remove from the list
 *
 * Return: Pointer to the next item.
 *
 **************************************************************************/

FileItem* RecentFiles::remove(FileItem *item)
{
    FileItem *next;

    next = item->next;
    if (item->prev)
        item->prev->next = next;
    if (item->next)
        item->next->prev = item->prev;

    free((void*)item->info.pathname);
    if (item->info.label)
	free((void*)item->info.label);
    free(item);

    return next;
}


/**************************************************************************
 *
 * Function: dump
 *
 * Description: Dumps the file list to stdout. This function is used for
 *	debugging purposes.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/
void RecentFiles::dump()
{
    FileItem *item = fileList;

    while (item) {
	printf("Pathname: %s\n", item->info.pathname);
	printf("Label: %s\n", (item->info.label)? item->info.label:
								"<default>");
	item = item->next;
    }
}


/**************************************************************************
 *
 * Function: readList
 *
 * Description: Reads the file list from the disk. Note that the previous
 *	contents of the file list, if any, are lost and are replaced with
 *	the contents of the disk file.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void RecentFiles::readList()
{
    FILE *fptr;
    FileInfo info;
    char buffer[MAX_FILELINE + 2];

    #ifdef DEBUG
    printf("Reading %s\n", diskPathname);
    #endif

    /*
     * First we open the file for reading and lock it with a shared
     * lock so others can read it to. If we cannot open the file we
     * silently blow off reading the list.
     */
    if ((fptr = fopen(diskPathname, "r")) == NULL)
	return;
    if (flock(fileno(fptr), LOCK_EX) == -1) {
	perror("flock");
	fclose(fptr);
	return;
    }

    /*
     * Read the version line. If it does not start with 'GVI'
     * blow off reading the rest of the file.
     */
    if (fgets(buffer, MAX_FILELINE, fptr) == NULL)
	goto readCleanup;
    if (strncmp(buffer, "GVI", 3))
	goto readCleanup;

    /*
     * Construct the filename list from the disk contents
     */
    while (fileList)
	fileList = remove(fileList);
    fileList = NULL;
    while (fgets(buffer, MAX_FILELINE, fptr)) {
	if (buffer[strlen(buffer) - 1] == '\n')
	    buffer[strlen(buffer) - 1] = '\0';
	if ((info.pathname = strtok(buffer, ":")) == NULL)
	    continue;
	info.label = strtok(NULL, ":");
	add(&info);
    }

    /*
     * Release the lock and close the file
     */
readCleanup:
    flock(fileno(fptr), LOCK_UN);
    fclose(fptr);
}


/**************************************************************************
 *
 * Function: writeList
 *
 * Description: Writes the file list to the disk.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void RecentFiles::writeList()
{
    int fd;
    FILE *fptr;

    #ifdef DEBUG
    printf("Writing %s\n", diskPathname);
    #endif

    /*
     * First we open the file in create mode. We then exclusively
     * lock the file and truncate it. We truncate after locking so
     * that noone reads an empty file. Finally we add a stream to the
     * file descriptor.
     */
    if ((fd = open(diskPathname, O_WRONLY | O_CREAT, 0666)) == -1) {
	perror("open");
	return;
    }
    if (flock(fd, LOCK_EX) == -1) {
	perror("flock");
	close(fd);
	unlink(diskPathname);
	return;
    }
    if (ftruncate(fd, 0) == -1) {
	perror("ftruncate");
	close(fd);
	unlink(diskPathname);
	return;
    }
    if ((fptr = fdopen(fd, "w")) == NULL) {
	fprintf(stderr, "fpopen failed\n");
	close(fd);
	unlink(diskPathname);
	return;
    }

    /*
     * Write the version number at the top of the file. Note that
     * the file MUST START with 'GVI' in the first column of the
     * first line. Otherwise the read routine will ignore the file.
     */
    fprintf(fptr, "GVI Version: %d\n", DISKLIST_VERSION);

    /*
     * Write the filename list
     */
    FileItem *item = fileList;
    while (item) {
	fprintf(fptr, "%s", item->info.pathname);
	if (item->info.label)
	    fprintf(fptr, ":%s", item->info.label);
	fprintf(fptr, "\n");
	item = item->next;
    }

    /*
     * Release the lock and close the file
     */
    flock(fd, LOCK_UN);
    fclose(fptr);
}


/**************************************************************************
 *
 * Function: getLabel
 *
 * Description: Returns the tab label for the specified pathname.
 *
 * Parameters: 
 *	pathname (I) - full pathname of file for which tab label is desired
 *
 * Return: Pointer to tab label or NULL if the default label is to be used
 *	or pathname cannot be found in the list.
 *
 **************************************************************************/

char* RecentFiles::getLabel(const char *pathname)
{
    FileItem *item = fileList;

    while (item) {
	if (strcmp(item->info.pathname, pathname) == 0)
	    return item->info.label;
	item = item->next;
    }
    return NULL;
}
