/**************************************************************************
 *
 * 		  Copyright (c)	1994 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: ViWsh.c++
 *
 * Description: VK component that provides a vi editor running in a forked
 *	window shell (wsh).
 *
 **************************************************************************/


#ident "$Revision: 1.2 $"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <limits.h>
#include <sys/types.h>
#include <unistd.h>
#include <Xm/BulletinB.h>
#include <Vk/VkApp.h>
#include <Vk/VkSimpleWindow.h>
#include <Vk/VkResource.h>
#include "ViWsh.h"
#include "safeMemory.h"
#include "common.h"
#include "ProgramMsg.h"


// Commands

#define XWSH_CMD	"/usr/sbin/xwsh"
#define VI_CMD		"/usr/bin/vi"


// Callback names

const char *const ViWsh::mappedCallback = "mappedCallback";
const char *const ViWsh::destroyedCallback = "destroyedCallback";


// Static class variables

int ViWsh::widthInc = 1;
int ViWsh::heightInc = 1;


/**************************************************************************
 *
 * Function: ViWsh
 *
 * Description: Constructor for the class.
 *
 * Parameters: 
 *	name (I) - name of the class instance
 *	parent (I) - parent widget for the instance
 *	filename (I) - name of file to edit. If NULL, then just a vi
 *			is run.
 *	width (I) - desired width in pixels
 *	height (I) - desired height in pixels
 *
 *	Note that the width and height will be used to calculate the
 *	number of columns and lines. Thus the actual width and height of
 *	the window may differ from that specified.
 *
 * Return: none
 *
 **************************************************************************/

ViWsh::ViWsh(const char *name, Widget parent, char *filename,
		Dimension width, Dimension height)
	:VkComponent(name)
{
    int numColumns = width / widthInc;
    int numLines = height / heightInc;
    buildIt(parent, filename, numColumns, numLines);
}


/**************************************************************************
 *
 * Function: ViWsh
 *
 * Description: Constructor for the class.
 *
 * Parameters: 
 *	name (I) - name of the class instance
 *	parent (I) - parent widget for the instance
 *	filename (I) - name of file to edit. If NULL, then just a vi
 *			is run.
 *
 * Return: none
 *
 **************************************************************************/

ViWsh::ViWsh(const char *name, Widget parent, char *filename)
    :VkComponent(name)
{
    int numColumns = (int)VkGetResource(parent,
					"viWsh.numColumns", "ViWsh.NumColumns",
					XtRCardinal, "80");
    int numLines = (int)VkGetResource(parent,
					"viWsh.numLines", "ViWsh.NumLines",
					XtRCardinal, "40");
    buildIt(parent, filename, numColumns, numLines);
}


/**************************************************************************
 *
 * Function: ~ViWsh
 *
 * Description: Destructor for the class.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

ViWsh::~ViWsh()
{
    if (pathname)
	free(pathname);
    if (wshName)
	free(wshName);
}


/**************************************************************************
 *
 * Function: className
 *
 * Description: Returns the name of the class.
 *
 * Parameters: none
 *
 * Return: Pointer to class name string.
 *
 **************************************************************************/

const char* ViWsh::className()
{
    return "ViWsh";
}


/**************************************************************************
 *
 * Function: buildIt
 *
 * Description: Called by the constructors to perform the actual work of
 *	setting up and running a viwsh.
 *
 * Parameters: 
 *	parent (I) - parent widget for the instance
 *	filename (I) - name of file to edit. If NULL, then just a vi
 *			is run.
 *	numColumns (I) - number of columns
 *	numLines (I) - number of lines
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::buildIt(Widget parent, char *filename, int numColumns, int numLines)
{
    char buffer[4096];
    char *baseFname, *minPathname;
    DEFARGS(10);

    /*
     * Instantiate a drawing area as our base widget.
     */
    STARTARGS;
    SETARG(XmNmarginWidth, 0);
    SETARG(XmNmarginHeight, 0);
    _baseWidget = XtCreateWidget(_name, xmBulletinBoardWidgetClass,
					parent, ARGLIST);

    /*
     * Install a destruction handler
     */
    installDestroyHandler();

    /*
     * Indicate that we are busy
     */
    theApplication->busy();

    /*
     * Init the resize accumulators
     */
    widthRem = heightRem = 0;

    /*
     * If a filename has been specified, we need the base filename
     * (i.e. path removed). We also want to construct and save for
     * later the absolute pathname of the file. And lastly we want
     * to construct the minimum pathname. This is a pathname with
     * the cwd stripped off. We use this to invoke vi so that we
     * minimize the chance of vi's status line running long and
     * requiring the user to hit return to actual edit the file.
     */
    if ((pathname = makeAbsolutePathname(filename)) == NULL) {
	baseFname = "noname";
        minPathname = NULL;
    } else {
        baseFname = basename(filename);
	minPathname = makeMinPathname(pathname);
    }
    #ifdef DEBUG
    printf("Filename = \"%s\"\n", filename);
    printf("Pathname = \"%s\"\n", pathname);
    printf("Base Filename = \"%s\"\n", baseFname);
    printf("minPathname = \"%s\"\n", minPathname);
    #endif

    /*
     * Make sure that the xwsh instance name is unique. Do this by taking
     * the file basename and appending on the pid.
     */
    sprintf(buffer, "%s-%d-%d", baseFname, this, getpid()); 
    wshName = safeStrdup(buffer);

    /*
     * Form the command to execute
     */
    sprintf(buffer, "%s -geometry %dx%d+%d+0 -name %s -title %s "
		"-nomenu -noscrollbar -vb "
		"-xrm \"XWsh*autoFork: False\" "
		"-xrm \"XWsh*allowSendEvents: True\" "
		"-xrm \"XWsh*overrideRedirect: True\" -e %s %s",
		XWSH_CMD, numColumns, numLines,
		XWidthOfScreen(XtScreen(_baseWidget)) + 10,
		wshName, baseFname, 
		VI_CMD, (minPathname)? minPathname: "");
    #ifdef DEBUG
    printf("%s\n", buffer);
    #endif

    /*
     * Execute the command
     */
    doReparent = False;
    wshFound = False;
    wshWid = NULL;
    VkProgram prog(buffer);
    prog.addCallback(VkProgram::exitCallback, this,
					(VkCallbackMethod)&ViWsh::viExitCB);
    wshProcess = prog.run();
}


/**************************************************************************
 *
 * Function: handleMapping
 *
 * Description: Processes an override redirect window mapping event. The
 *	function determines if this window is our wsh client and if so
 *	records the relevant information and supervises the window
 *	reparenting.
 *
 * Parameters: 
 *	wid (I) - X window ID of mapped window
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::handleMapping(Window wid)
{
    XClassHint xch;

    /*
     * If we are looking for a window, determine if this is the one we
     * want by looking at the window mgr instance name.
     */
    if (wshFound)
	return;
    if (XGetClassHint(XtDisplay(_baseWidget), wid, &xch)) {
	if (!strcmp(xch.res_name, wshName)) {
	    wshFound = True;
	    wshWid = wid;
	}
	XFree(xch.res_name);
	XFree(xch.res_class);
    }

    /*
     * If we found the window, reparent it.
     */
    if (wshFound) {
	/*
	 * Set our base widget's size based on the wsh window
	 */
	setSize();

	/*
	 * If the base widget is not realized defer the reparenting until
	 * it is realized.
	 */
	if (XtIsRealized(_baseWidget)) {
	    doReparent = False;
	    reparentWindow();
	} else {
	    doReparent = True;
	}

	/*
	 * Call the user's mapped callbacks
	 */
        callCallbacks(mappedCallback, (void*)wshWid);

	/*
	 * Indiacte we are no longer busy
	 */
	theApplication->notBusy();
    }
}


/**************************************************************************
 *
 * Function: handleResize
 *
 * Description: Resizes the ViWsh based on the size change specified.
 *
 * Parameters: 
 *	dw (I) - change in width
 *	dh (I) - chane in height
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::handleResize(int dw, int dh)
{
    DEFARGS(10);
    XWindowAttributes xwa;
    int w, h;
    Dimension width, height;

    /*
     * Resize the viwsh window and its parent. Note that we must
     * ensure that the viwsh window is only resized by its native
     * width and height resize increments or we can get a bit of
     * movement of the viwsh window relative to its parent.
     */
    XGetWindowAttributes(XtDisplay(_baseWidget), wshWid, &xwa);
    widthRem += dw;
    heightRem += dh;
    w = xwa.width;
    h = xwa.height;
    if (widthRem < 0) {
	w -= widthInc;
	widthRem += widthInc;
    }
    if (widthRem / widthInc != 0) {
	w += widthRem / widthInc * widthInc;
	widthRem %= widthInc;
    }
    if (heightRem < 0) {
	h -= heightInc;
	heightRem += heightInc;
    }
    if (heightRem / heightInc != 0) {
	h += heightRem / heightInc * heightInc;
	heightRem %= heightInc;
    }
    XResizeWindow(XtDisplay(_baseWidget), wshWid, w, h);

    /*
     * Now resize the base widget
     */
    STARTARGS;
    SETARG(XmNwidth, &width);
    SETARG(XmNheight, &height);
    XtGetValues(_baseWidget, ARGLIST);
    STARTARGS;
    SETARG(XmNwidth, width + dw);
    SETARG(XmNheight, height + dh);
    XtSetValues(_baseWidget, ARGLIST);
}


/**************************************************************************
 *
 * Function: afterRealizeHook
 *
 * Description: If the base widget is not realized and we need to do
 *	reparenting, do it now.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::afterRealizeHook()
{
    if (doReparent) {
	doReparent = False;
	reparentWindow();
    }
}


/**************************************************************************
 *
 * Function: setSize
 *
 * Description: Sets the base widget's size based on the size of the
 *	vi/wsh window.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::setSize()
{
    XWindowAttributes xwa;
    DEFARGS(10);

    ASSERT(wshWid != NULL);

    /*
     * Get the wsh window's dimensions
     */
    XGetWindowAttributes(XtDisplay(_baseWidget), wshWid, &xwa);

    /*
     * Set the base widget dimensions to the wsh dims.
     */
    STARTARGS;
    SETARG(XmNheight, xwa.height);
    SETARG(XmNwidth, xwa.width);
    XtSetValues(_baseWidget, ARGLIST);
}


/**************************************************************************
 *
 * Function: reparentWindow
 *
 * Description: Performs the reparenting of the wsh window into our
 *	window hierarchy.
 *
 * Parameters: none
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::reparentWindow()
{
    XWindowAttributes xwa;
    XSizeHints wshHints;
    long flags;

    /*
     * Get the wsh window's dimensions
     */
    XGetWindowAttributes(XtDisplay(_baseWidget), wshWid, &xwa);

    /*
     * Reparent the viwsh window to the base widget
     */
    XReparentWindow(XtDisplay(_baseWidget), wshWid, XtWindow(_baseWidget),
									0, 0);

    /*
     * We need to be aware of the resizing increment required by the
     * wsh window so that we resize appropriately. We should also set
     * an app minimum size if that is what the viwsh requires.
     */
    XGetWMNormalHints(XtDisplay(_baseWidget), wshWid, &wshHints, &flags);
    if ((flags & PResizeInc) && wshHints.width_inc && wshHints.height_inc) {
	widthInc = wshHints.width_inc;
	heightInc = wshHints.height_inc;
    }
}


/**************************************************************************
 *
 * Function: viExitCB
 *
 * Description: Called when a vi/wsh window process exits.
 *
 * Parameters:
 *      obj (I) - component that triggered the callback
 *      clientData (I) - not used
 *      callData (I) - callback data
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::viExitCB(VkCallbackObject*, void*, void*)
{
    (void)XtAppAddTimeOut(theApplication->appContext(), 0,
						viExitTimerCB, this);
}


/**************************************************************************
 *
 * Function: viExitTimerCB
 *
 * Description: Called when a zero length timer completes. The timer is
 *	set in the viExitCB routine. This is to sync up with the X
 *	command stream.
 *
 * Parameters: 
 *	clientData (I) - class instance pointer
 *	idp (I) - ID of interval timer. Not used.
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::viExitTimerCB(XtPointer clientData, XtIntervalId*)
{
    ViWsh *viWsh = (ViWsh*)clientData;
    viWsh->callCallbacks(destroyedCallback, (void*)NULL);
}


/**************************************************************************
 *
 * Function: sendString
 *
 * Description: Sends a NULL terminated string to the reparented vi/wsh
 *	window.
 *
 * Parameters: 
 *	str (I) - string to send
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::sendString(const char *str)
{
    if (wshWid) {
	while (*str != '\0')
	    sendChar(*str++);
	XFlush(XtDisplay(_baseWidget));
    }
}


/**************************************************************************
 *
 * Function: sendChar
 *
 * Description: Sends a character to the reparented vi/wsh window by
 *	simulating a keypress/keyrelease event. Wildly ripped off from
 *	CaseVision's ExecutionView.
 *
 * Parameters: 
 *	ch (I) - character to send to vi.
 *
 * Return: none
 *
 **************************************************************************/

void ViWsh::sendChar(char ch)
{
    XEvent ev;
    char buf[256];
    KeySym keysym;
    KeySym generatedKeysym;
    unsigned int generatedModifiers;

    if (wshWid == NULL)
	return;

    ev.type = KeyPress;
    ev.xkey.window = wshWid;
    ev.xkey.subwindow = wshWid;
    ev.xkey.display = XtDisplay(_baseWidget);
    ev.xkey.same_screen = True;
    ev.xkey.root = XRootWindowOfScreen(XtScreen(_baseWidget));
    ev.xkey.x = 0;
    ev.xkey.y = 0;
    // We should really translate these coordinates...
    ev.xkey.x_root = 0;
    ev.xkey.y_root = 0;
    ev.xkey.state = 0;

    if (('a' <= ch && ch <= 'z') || ('0' <= ch && ch <= '9')){
	buf[0] = ch;
	buf[1] = '\0';
    } else if ('A' <= ch && ch <= 'Z'){
	//ev.xkey.state=ShiftMask;
	buf[0] = ch;
	buf[1] = '\0';
    } else {
	switch (ch) {
	    case '\033':
		strcpy(buf,"Escape");
		break;
	    case '\t':
		strcpy(buf,"Tab");
		break;
    	    case '\n':
      		strcpy(buf,"Return");
      		break;
    	    case ' ':
      		strcpy(buf,"space");
      		break;
    	    case '!':
      		strcpy(buf,"exclam");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '\"':
      		strcpy(buf,"quotedbl");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '#':
      		strcpy(buf,"numbersign");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '$':
      		strcpy(buf,"dollar");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '%':
      		strcpy(buf,"percent");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '&':
      		strcpy(buf,"ampersand");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '\'':
      		strcpy(buf,"apostrophe");
      		break;
	    /*
	    case '\'':
      		strcpy(buf,"quoteright");
      		break;
	    */
	    case '\(':
      		strcpy(buf,"parenleft");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '\)':
      		strcpy(buf,"parenright");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '*':
      		strcpy(buf,"asterisk");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '+':
      		strcpy(buf,"plus");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case ',':
      		strcpy(buf,"comma");
      		break;
    	    case '-':
      		strcpy(buf,"minus");
      		break;
    	    case '.':
      		strcpy(buf,"period");
      		break;
    	    case '/':
      		strcpy(buf,"slash");
      		break;
    	    case ':':
      		strcpy(buf,"colon");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case ';':
      		strcpy(buf,"semicolon");
      		break;
    	    case '<':
      		strcpy(buf,"less");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '=':
      		strcpy(buf,"equal");
      		break;
    	    case '>':
      		strcpy(buf,"greater");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '?':
      		strcpy(buf,"question");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '\[':
      		strcpy(buf,"bracketleft");
      		break;
    	    case '\\':
      		strcpy(buf,"backslash");
      		break;
    	    case '\]':
      		strcpy(buf,"bracketright");
      		break;
    	    case '^':
      		strcpy(buf,"asciicircum");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '_':
      		strcpy(buf,"underscore");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '@':
      		strcpy(buf,"at");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '`':
      		strcpy(buf,"quoteleft");
      		break;
    	    case '\{':
      		strcpy(buf,"braceleft");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '|':
      		strcpy(buf,"bar");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '\}':
      		strcpy(buf,"braceright");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    case '~':
      		strcpy(buf,"asciitilde");
      		//ev.xkey.state=ShiftMask;
      		break;
    	    default:
      		buf[0] = ch;
      		buf[1] = '\0';
      		break;
        }
    }
    keysym = XStringToKeysym(buf);
    if (keysym == NoSymbol) {
	progMsg.error("not a valid keysym: %s", buf);
    }
    if (!(ev.xkey.keycode = XKeysymToKeycode(XtDisplay(_baseWidget),
								keysym))) {
	progMsg.error("not a valid keycode: %s", buf);
    }

    ev.xkey.state = 0;
    XtTranslateKeycode(XtDisplay(_baseWidget), ev.xkey.keycode,
				ev.xkey.state,
				&generatedModifiers, &generatedKeysym);
    if (keysym != generatedKeysym) {
	ev.xkey.state = ShiftMask;
	XtTranslateKeycode(XtDisplay(_baseWidget), ev.xkey.keycode,
			ev.xkey.state, &generatedModifiers,
			&generatedKeysym);
	if (keysym != generatedKeysym) {
	    ev.xkey.state = LockMask;
	    XtTranslateKeycode(XtDisplay(_baseWidget), ev.xkey.keycode,
				ev.xkey.state, &generatedModifiers,
				&generatedKeysym);
	    if (keysym != generatedKeysym) {
		ev.xkey.state = ControlMask;
		XtTranslateKeycode(XtDisplay(_baseWidget), ev.xkey.keycode,
				        ev.xkey.state,
					&generatedModifiers,
					&generatedKeysym);
		if (keysym != generatedKeysym) {
		    ev.xkey.state = 0;
		}
	    }
	}
    }

    XSendEvent(ev.xkey.display, ev.xkey.window, True, KeyPressMask, &ev);
    ev.type = KeyRelease;
    XSendEvent(ev.xkey.display, ev.xkey.window, True, KeyReleaseMask, &ev);
}


/**************************************************************************
 *
 * Function: makeMinPathname
 *
 * Description: Constructs a minimum pathname from the specified pathname.
 *	A minimum pathname is a pathname with the cwd stripped off if the
 *	original pathname begins with the cwd.
 *
 * Parameters: 
 *	path (I) - pathname to minimize
 *
 * Return: Pointer to minimum pathname. This is a pointer into the original
 *	pathname's storage so it should not be freed.
 *
 **************************************************************************/

char* ViWsh::makeMinPathname(char *path)
{
    char *cwd, *minPath;
    size_t len;
    int needFree = 0;

    if ((cwd = getenv("PWD")) == NULL) {
        if ((cwd = getcwd(NULL, PATH_MAX + 1)) == NULL)
	    return path;
	needFree = 1;
    }
    len = strlen(cwd);
    if (strncmp(cwd, path, len) == 0) {
	minPath = path + len;
	if (*minPath == '/')
	    minPath++;
    } else
	minPath = path;
    if (needFree)
        free(cwd);

    return minPath;
}
