/**************************************************************************
 *
 * 		  Copyright (c)	1993 Silicon Graphics, Inc.
 *			All Rights Reserved
 *
 *	   THIS	IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SGI
 *
 * The copyright notice above does not evidence any actual of intended
 * publication of such source code, and is an unpublished work by Silicon
 * Graphics, Inc. This material contains CONFIDENTIAL INFORMATION that is
 * the property of Silicon Graphics, Inc. Any use, duplication or
 * disclosure not specifically authorized by Silicon Graphics is strictly
 * prohibited.
 *
 * RESTRICTED RIGHTS LEGEND:
 *
 * Use, duplication or disclosure by the Government is subject to
 * restrictions as set forth in subdivision (c)(1)(ii) of the Rights in
 * Technical Data and Computer Software clause at DFARS 52.227-7013,
 * and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 * Supplement. Unpublished - rights reserved under the Copyright Laws of
 * the United States. Contractor is SILICON GRAPHICS, INC., 2011 N.
 * Shoreline Blvd., Mountain View, CA 94039-7311
 **************************************************************************
 *
 * File: safeMemory.c++
 *
 * Description: Safe memory allocation functions. These functions check
 *	the return value of memory allocation functions. If the memory
 *	allocation fails, a message is printed and the program terminates.
 *
 **************************************************************************/


#ident "$Revision: 1.1 $"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "ProgramMsg.h"
#include "safeMemory.h"


/**************************************************************************
 *
 * Function: safeMalloc
 *
 * Description: Provides a safety wrapper around malloc. If malloc fails,
 *	a message is printed and exit is called.
 *
 * Parameters: 
 *	size (I) - number of bytes of memory to malloc.
 *
 * Return: Pointer to allocated memory.
 *
 **************************************************************************/

void* safeMalloc(size_t size)
{
    ASSERT(size != 0);
    register void *mptr = malloc(size);
    if (mptr == NULL) {
	progMsg.error("malloc failed while allocating %d bytes of memory",
									size);
	exit(1);
    }
    return mptr;
}


/**************************************************************************
 *
 * Function: safeCalloc
 *
 * Description: Provides a safety wrapper around calloc. If calloc fails,
 *	a message is printed and exit is called.
 *
 * Parameters: 
 *	nelem (I) - number of elements to allocate
 *	elsize (I) - number of bytes in each element
 *
 * Return: Pointer to allocated memory.
 *
 **************************************************************************/

void* safeCalloc(size_t nelem, size_t elsize)
{
    ASSERT((nelem != 0) && (elsize != 0));
    register void *mptr = calloc(nelem, elsize);
    if (mptr == NULL) {
	progMsg.error("calloc failed while allocating %d bytes of memory",
							nelem * elsize);
	exit(1);
    }
    return mptr;
}


/**************************************************************************
 *
 * Function: safeRealloc
 *
 * Description: Provides a safety wrapper around realloc. If realloc fails,
 *	a message is printed and exit is called.
 *
 * Parameters: 
 *	ptr (I) - pointer to original memeory block
 *	size (I) - new size for memory block
 *
 * Return: Pointer to resized memory.
 *
 **************************************************************************/

void* safeRealloc(void *ptr, size_t size)
{
    ASSERT((ptr != NULL) && (size != 0));
    register void *mptr = realloc(ptr, size);
    if (mptr == NULL) {
	progMsg.error("realloc failed while allocating %d bytes of memory",
									size);
	exit(1);
    }
    return mptr;
}


/**************************************************************************
 *
 * Function: safeStrdup
 *
 * Description: Provides a safety wrapper around strdup. If strdup fails,
 *	a message is printed and exit is called.
 *
 * Parameters: 
 *	str (I) - string to duplicate
 *
 * Return: Pointer to stroage containing the copy of the string.
 *
 **************************************************************************/

char* safeStrdup(const char *str)
{
    ASSERT(str != NULL);
    register char *cptr = strdup(str);
    if (cptr == NULL) {
	progMsg.error("strdup failed while allocating %d bytes of memory",
							strlen(str) + 1);
	exit(1);
    }
    return cptr;
}
